/*
 * Copyright (c) 1995, 1998 Hellmuth Michaelis. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *---------------------------------------------------------------------------*
 *
 *	LED display driver - example of a working LKM driver
 *	----------------------------------------------------
 *
 *	Hellmuth Michaelis, hm@kts.org
 *
 *	last edit-date: [Thu Nov  5 11:14:38 1998]
 *
 *---------------------------------------------------------------------------
 *
 *	This driver is for a home-made 2-digit LED 7 segment display 
 *	controlled by a 8255 PIO which is used to replace the spoiler
 *	speed LED display found on many pc cases with an external
 *	controllable one.
 *
 *	8255 port A is connected to the lsb (right) 7 segment LED
 *	8255 port B is connected to the msb (left) 7 segment LED
 *
 *	Wiring is straightforward, the 8255 ports are connected via a
 *	non inverting TTL open collector driver to the LED pins, data
 *	bit 0 to LED pin a, bit 1 to pin b and so on.
 *
 *	8255 port C is not used and configured as an input.
 *
 *	The device accepts input from 2 minor numbered devices: device
 *	minor 0 will accept ASCII characters ' ' (space) = blank digit,
 *	'0'...'9' and 'a'...'f' (also 'A'...'F'). These characters will
 *	be translated to 7segment values and displayed on the display.
 *	Device minor number 1 will accept any binary value which is sent
 *	untranslated to the display to display anything on the 7segment
 *	LED.
 *
 *	For both devices, the first value sent will be displayed on the
 *	right (lsb) LED. The next write will shift the right LED value
 *	to the left (msb) LED and the new value will be displayed in the
 *	right (lsb) LED and so on.
 *
 *---------------------------------------------------------------------------*/

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/exec.h>
#include <sys/conf.h>
#include <sys/sysent.h>
#include <sys/lkm.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/sysproto.h>

/* the base address of the 8255 pio, hardwired for LKM */

#define BASEADDR	0x200		/* 8255 base address */

/* the 8255 port offsets */

#define PORTA		(BASEADDR+0)	/* 8255 port A */
#define PORTB		(BASEADDR+1)	/* 8255 port B */
#define PORTC		(BASEADDR+2)	/* 8255 port C */
#define PORTCNTL	(BASEADDR+3)	/* 8255 control port */

/* the config value for 8255 port usage initialization */

#define INIT8255	0x89		/* port a+b = output, c = input */

/* blank the diaplay */

#define	BLANK		0x00		/* blank display value */

/* forward declarations */

static void led_init(void);
static void led_blank(void);
static void led_display(unsigned char value, int xlat_flag);
int ledopen(int dev, int flag, int mode, struct proc *p);
int ledclose(int dev, int flag, int mode, struct proc *p);
int ledwrite(dev_t dev, struct uio *uio);

/* character device table entry, this gets copied at modload time into	*/
/* the kernels lkm dummy device driver entries, -> sys/i386/i386/conf.c */

struct cdevsw dev_led = {
	(d_open_t *)	ledopen,	/* open		*/
	(d_close_t *)	ledclose,	/* close	*/
	(d_read_t *)	nxread,		/* read		*/
	(d_write_t *)	ledwrite,	/* write	*/
  	(d_ioctl_t *)	nxioctl,	/* ioctl	*/
	(d_stop_t *)	nxstop,		/* stop		*/
	(d_reset_t *)	nxreset,	/* reset	*/
	(d_devtotty_t *)nxdevtotty,	/* ttys		*/
  	(d_select_t *)	nxselect,	/* select	*/
	(d_mmap_t *)	nxmmap,		/* mmap		*/
	(d_strategy_t *)nxstrategy,	/* strategy	*/
};

/* construct lkm_dev structure, see: lkm.h */

MOD_DEV (led, LM_DT_CHAR, -1, &dev_led);

/*---------------------------------------------------------------------------*
 *	function called when (mod-) loading the driver
 *---------------------------------------------------------------------------*/
int led_load(struct lkm_table *lkmtp, int cmd)
{
	uprintf("LED driver: led_load called, driver installed\n");
	led_init();	/* init port */
	led_blank();	/* blank led's */
	return 0;
}

/*---------------------------------------------------------------------------*
 *	function called when (mod-) unloading the driver
 *---------------------------------------------------------------------------*/
int led_unload(struct lkm_table *lkmtp, int cmd)
{
	led_blank();	/* blank led's */	
	uprintf("LED driver: led_unload called, driver removed\n");
	return 0;
}

/*---------------------------------------------------------------------------*
 *	dispatcher function for (mod-) load/unload/stat
 *---------------------------------------------------------------------------*/
int led_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
	DISPATCH(lkmtp, cmd, ver, led_load, led_unload, lkm_nullcmd);
}

/*---------------------------------------------------------------------------*
 *	open device
 *---------------------------------------------------------------------------*/
int ledopen(int dev, int flag, int mode, struct proc *p)
{
	if((minor(dev) < 0) || (minor(dev) > 1))
		return (ENXIO);
	return(0);
}

/*---------------------------------------------------------------------------*
 *	close device
 *---------------------------------------------------------------------------*/
int ledclose(int dev, int flag, int mode, struct proc *p)
{
	return(0);
}

/*---------------------------------------------------------------------------*
 *	write to device
 *---------------------------------------------------------------------------*/
int ledwrite(dev_t dev, struct uio *uio)
{
	unsigned char ledbyte;
	int error;

	/* get byte from user space into kernel space */
	
	if((error = uiomove(&ledbyte, 1, uio)) != 0)
		return(error);

	if((minor(dev)) == 0)		/* ascii -> 7segment xlate */
		led_display(ledbyte, 1);
	else				/* real hardware bits */
		led_display(ledbyte, 0);
	return(0);
}

/*---------------------------------------------------------------------------*
 *	init the 8255
 *---------------------------------------------------------------------------*/
static void led_init(void)
{
	outb(PORTCNTL, INIT8255);	/* initialize value for 8255 */
}

/*---------------------------------------------------------------------------*
 *	blank both displays
 *---------------------------------------------------------------------------*/
static void led_blank(void)
{
	outb(PORTA, BLANK);		/* blank lsb */
	outb(PORTB, BLANK);		/* blank msb */
}

/*---------------------------------------------------------------------------*
 *	write to the LED's
 *---------------------------------------------------------------------------*/
static void led_display(unsigned char value, int xlat_flag)
{
	static unsigned char lo_led = 0;
	static unsigned char hi_led = 0;

	unsigned char output;
	
	/* translation table bin -> 7segment */

	static unsigned char int2led[] =
	{
		0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07,
		0x7f, 0x6f, 0x77, 0x7c, 0x39, 0x5e, 0x79, 0x71
	};

	if(xlat_flag)			/* translated output */
	{
		if(value == ' ')			/* blank */
			output = 0;
		else if(value >= '0' && value <= '9')	/* number 0-9 */
			output = int2led[value -= '0'];
		else if(value >= 'a' && value <= 'f')	/* char a-f */
			output = int2led[(value - 'a') + 10];
		else if(value >= 'A' && value <= 'F')	/* char a-f */
			output = int2led[(value - 'A') + 10];
		else
			return;				/* invalid */
	}
	else			/* no translation */
	{
		output = value;		/* 1 to 1 */
	}

	hi_led = lo_led;		/* shift lsb -> msb */
	lo_led = output;		/* lsb = new value */

	outb(PORTA, lo_led);		/* output msb */
	outb(PORTB, hi_led);		/* output lsb */
}

/*-------------------------------- EOF ---------------------------------------*/
